#include <errno.h>
#include <fcntl.h>
#include <libintl.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>

#define SYSLOG_NAMES
#include <syslog.h>

#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <sys/mman.h>

#define _(String) gettext((String))

#include <popt.h>

#include <regex.h>

#include "initlog.h"
#include "process.h"
#include "alloc.h"

static int logfacility = LOG_DAEMON;
static int logpriority = LOG_NOTICE;
static int reexec = 0;
static int quiet = 0;
static int debug = 0;
static int logEntries = 0;
static struct logInfo *logData = 0;

regex_t **regList = 0;

static int
parse_integer(const char *s, int *res)
{
	unsigned long ul;
	char   *ep;

	ul = strtoul(s, &ep, 10);
	if (ul > INT_MAX || *ep != '\0')
		return -1;

	*res = ul;
	return 0;
}

static void
get_facility(const char *s, int *fa)
{
	if (s && *s && parse_integer(s, fa) < 0)
	{
		unsigned i;

		*fa = LOG_DAEMON;
		for (i = 0; facilitynames[i].c_name; ++i)
		{
			if (!strcasecmp(s, facilitynames[i].c_name))
			{
				*fa = facilitynames[i].c_val;
				break;
			}
		}
	}
}

static void
get_priority(const char *s, int *pri)
{
	if (s && *s && parse_integer(s, pri) < 0)
	{
		unsigned i;

		*pri = LOG_NOTICE;
		for (i = 0; prioritynames[i].c_name; ++i)
		{
			if (!strcasecmp(s, prioritynames[i].c_name))
			{
				*pri = prioritynames[i].c_val;
				break;
			}
		}
	}
}

static void
readConfiguration(const char *fname)
{
	int     fd = open(fname, O_RDONLY);

	if (fd < 0)
		return;
	else
	{
		struct stat sbuf;

		if (fstat(fd, &sbuf))
		{
			close(fd);
			return;
		} else
		{
			char   *data = (char *) mmap(0, 1 + sbuf.st_size,
						     PROT_READ | PROT_WRITE,
						     MAP_PRIVATE, fd, 0);

			close(fd);
			if ((void *) data == MAP_FAILED)
				return;
			else
			{
				unsigned num = 0;
				char   *line;

				while ((line = getLine(&data)))
				{
					const char *pattern;
					unsigned offset;

					if (line[0] == '#')
						continue;

					pattern = "ignore ";
					offset = sizeof("ignore ") - 1;
					if (!strncmp(line, pattern, offset))
					{
						regex_t *regexp =
							xmalloc(sizeof
								(regex_t));
						if (!regcomp
						    (regexp, line + offset,
						     REG_EXTENDED | REG_NOSUB
						     | REG_NEWLINE))
						{
							regList =
								xrealloc
								(regList,
								 (num +
								  2) *
								 sizeof
								 (regex_t *));
							regList[num] = regexp;
							regList[++num] = 0;
						}
						continue;
					}

					pattern = "facility ";
					offset = sizeof("facility ") - 1;
					if (!strncmp(line, pattern, offset))
					{
						get_facility(line + offset,
							     &logfacility);
						continue;
					}

					pattern = "priority ";
					offset = sizeof("priority ") - 1;
					if (!strncmp(line, pattern, offset))
					{
						get_priority(line + offset,
							     &logpriority);
						continue;
					}
				}
			}
		}
	}
}

char   *
getLine(char **data)
{
	/* Get one line from data */
	/* Anything up to a carraige return (\r) or a backspace (\b) is discarded. */
	/* If this really bothers you, mail me and I might make it configurable. */
	/* It's here to avoid confilcts with fsck's progress bar. */
	char   *x, *y = 0;

	if (!data || !*data)
		return 0;
	x = *data;

	while (*x && (*x != '\n'))
	{
		while (*x && (*x != '\n') && (*x != '\r') && (*x != '\b'))
			++x;
		if (*x && (*x == '\r' || *x == '\b'))
			*data = ++x;
	}

	if (*x)
	{
		y = xmalloc(++x - *data);
		memcpy(y, *data, x - *data - 1);
		y[x - *data - 1] = '\0';
	} else
	{
		if (x - *data)
		{
			y = xmalloc(x - *data + 1);
			y[x - *data - 1] = '\n';
			memcpy(y, *data, x - *data);
			y[x - *data] = '\0';
		}
		x = 0;
	}

	*data = x;
	return y;
}

static int
startDaemon()
{
	int     pid = fork();

	if (pid < 0)
	{
		error(EXIT_SUCCESS, errno, "fork");
		return -1;
	}

	if (pid)
	{
		/* parent */
		int     rc;

		if (waitpid(pid, &rc, 0) < 0)
		{
			error(EXIT_SUCCESS, errno, "waitpid");
			return -1;
		}

		if (WIFEXITED(rc))
		{
			DDEBUG("minilogd returned %d!\n", WEXITSTATUS(rc));
			return WEXITSTATUS(rc);
		} else if (WIFSIGNALED(rc))
		{
			DDEBUG("minilogd received signal %d!\n",
			       WTERMSIG(rc));
			return 128 + WTERMSIG(rc);
		} else
			return -1;
	} else
	{
		/* child */
		int     i, fd = open(_PATH_DEVNULL, O_RDWR);

		if (fd < 0)
			error(EXIT_FAILURE, errno, "open: %s", _PATH_DEVNULL);

		for (i = 0; i <= 2; ++i)
			if (fd != i)
				dup2(fd, i);
		if (fd > 2)
			close(fd);

		execl("/sbin/minilogd", "/sbin/minilogd", NULL);
		error(-1, errno, "execl");
		exit(-1);
	}
}

int
trySocket()
{
	int     s = socket(AF_LOCAL, SOCK_DGRAM, 0);

	if (s < 0)
		return 1;
	else
	{
		int     status;
		struct sockaddr_un addr;

		memset(&addr, 0, sizeof(addr));
		addr.sun_family = AF_LOCAL;
		strncpy(addr.sun_path, _PATH_LOG, sizeof(addr.sun_path) - 1);

		status = connect(s, (struct sockaddr *) &addr, sizeof(addr));
		close(s);
		if (status < 0)
		{
			if (EPROTOTYPE == errno)
			{
				DDEBUG("connect failed (EPROTOTYPE), trying stream\n");
				s = socket(AF_LOCAL, SOCK_STREAM, 0);
				status = connect(s,
						 (struct sockaddr *) &addr,
						 sizeof(addr));
				close(s);
				if (status < 0)
				{
					DDEBUG("connect failed: %s\n",
					       strerror(errno));
					return 1;
				}
				return 0;
			}
			DDEBUG("connect failed: %s\n", strerror(errno));
			return 1;
		}
	}
	return 0;
}

static int
line_is_empty(const char *line)
{
	if (!line)
		return 1;

	while (*line && isspace(*line))
		++line;

	return !*line;
}

int
logLine(struct logInfo *logEnt)
{
	/* Don't log empty or null lines */
	if (!line_is_empty(logEnt->line))
	{
		/* Logs a line... somewhere. */
		struct stat statbuf;

		if (((stat(_PATH_LOG, &statbuf) < 0) || trySocket())
		    && startDaemon())
		{
			DDEBUG("starting daemon failed, pooling entry %d\n",
			       logEntries);
			logData =
				xrealloc(logData,
					 (logEntries +
					  1) * sizeof(struct logInfo));
			logData[logEntries] = (*logEnt);
			++logEntries;
		} else
		{
			if (logEntries > 0)
			{
				int     i;

				for (i = 0; i < logEntries; ++i)
				{
					DDEBUG("flushing log entry %d =%s=\n",
					       i, logData[i].line);
					openlog(logData[i].cmd, 0,
						logData[i].fac);
					syslog(logData[i].pri, "%s",
					       logData[i].line);
					closelog();
				}
				free(logData);
				logEntries = 0;
			}
			DDEBUG("logging =%s= via syslog\n", logEnt->line);
			openlog(logEnt->cmd, 0, logEnt->fac);
			syslog(logEnt->pri, "%s", logEnt->line);
			closelog();
		}
	}
	return 0;
}

static void
fill_entry(struct logInfo *logentry, const char *cmd)
{
	if (cmd)
	{
		logentry->cmd = xstrdup(basename(cmd));
		if ((logentry->cmd[0] == 'K' || logentry->cmd[0] == 'S')
		    && isdigit(logentry->cmd[1]) && isdigit(logentry->cmd[2]))
			logentry->cmd += 3;
	} else
		logentry->cmd = xstrdup(_("(none)"));

	logentry->pri = logpriority;
	logentry->fac = logfacility;
}

int
logEvent(const char *cmd, int eventtype, const char *string)
{
	char   *eventtable[] = {
		_("%s babbles incoherently"),
		_("%s succeeded"),
		_("%s failed"),
		_("%s cancelled at user request"),
		_("%s failed due to a failed dependency")
			/* insert more here */
	};
	unsigned len;
	struct logInfo logentry;

	fill_entry(&logentry, cmd);

	if (!string)
		string = xstrdup(cmd);

	if (eventtype >= sizeof(eventtable) / sizeof(*eventtable))
		eventtype = 0;

	len = strlen(eventtable[eventtype]) + strlen(string);
	logentry.line = xmalloc(len);
	snprintf(logentry.line, len, eventtable[eventtype], string);

	return logLine(&logentry);
}

int
logString(const char *cmd, const char *string)
{
	struct logInfo logentry;

	fill_entry(&logentry, cmd);

	logentry.line = xstrdup(string);

	return logLine(&logentry);
}

int
processArgs(int argc, const char **argv, int silent)
{
	char   *cmdname = 0;
	char   *conffile = 0;
	char   *cmd = 0;
	char   *logstring = 0;
	char   *fac = 0, *pri = 0;
	int     cmdevent = 0;
	int     rc;
	poptContext context;
	struct poptOption optTable[] = {
		POPT_AUTOHELP {"conf", 0, POPT_ARG_STRING, &conffile, 0,
			       "configuration file (default: /etc/initlog.conf)",
			       0},
		{"name", 'n', POPT_ARG_STRING, &cmdname, 0,
		 "name of service being logged", 0},
		{"event", 'e', POPT_ARG_INT, &cmdevent, 0,
		 "event being logged (see man page)", 0},
		{"cmd", 'c', POPT_ARG_STRING, &cmd, 0,
		 "command to run, logging output", 0},
		{"debug", 'd', POPT_ARG_NONE, &debug, 0,
		 "print lots of verbose debugging info", 0},
		{"run", 'r', POPT_ARG_STRING, &cmd, 3,
		 "command to run, accepting input on open fd", 0},
		{"string", 's', POPT_ARG_STRING, &logstring, 0,
		 "string to log", 0},
		{"facility", 'f', POPT_ARG_STRING, &fac, 1,
		 "facility to log at (default: 'local7')", 0},
		{"priority", 'p', POPT_ARG_STRING, &pri, 2,
		 "priority to log at (default: 'notice')", 0},
		{"quiet", 'q', POPT_ARG_NONE, &quiet, 0,
		 "suppress stdout/stderr", 0},
		{0, 0, 0, 0, 0, 0}
	};

	context = poptGetContext("initlog", argc, argv, optTable, 0);

	while ((rc = poptGetNextOpt(context)) > 0)
	{
		switch (rc)
		{
			case 1:
				get_facility(fac, &logfacility);
				break;
			case 2:
				get_priority(pri, &logpriority);
				break;
			case 3:
				reexec = 1;
				break;
			default:
				break;
		}
	}

	if (rc < -1)
	{
		if (!silent)
			error(EXIT_SUCCESS, 0, "%s: %s",
			      poptBadOption(context,
					    POPT_BADOPTION_NOALIAS),
			      poptStrerror(rc));
		return -1;
	}

	if (cmd && logstring)
	{
		if (!silent)
			error(EXIT_SUCCESS, 0, "%s",
			      _
			      ("--cmd and --run are incompatible with --string"));
		return -1;
	}

	if (cmdname && (!logstring && !cmdevent && !cmd))
	{
		if (!silent)
			error(EXIT_SUCCESS, 0, "%s",
			      _
			      ("--name requires one of --event, --string or --cmd"));
		return -1;
	}

	if (cmdevent && (!cmdname && !logstring))
	{
		if (!silent)
			error(EXIT_SUCCESS, 0, "%s",
			      _
			      ("--event requires one of --cmdname or --logstring"));
		return -1;
	}

	readConfiguration(conffile ? : "/etc/initlog.conf");

	if (cmd)
		while (isspace(*cmd))
			++cmd;

	if (cmdevent)
		return logEvent(cmdname, cmdevent, logstring);
	else if (logstring)
		return logString(cmdname, logstring);
	else if (cmd && *cmd)
		return runCommand(cmdname, cmd, reexec, quiet, debug);
	else
	{
		if (!silent)
			error(EXIT_SUCCESS, 0, "%s", _("nothing to do"));
		return -1;
	}
	return 0;
}

int
main(int argc, const char *argv[])
{
#if 0
	setlocale(LC_ALL, "");
	bindtextdomain("initlog", "/etc/locale");
	textdomain("initlog");
#endif
	exit(processArgs(argc, argv, 0));
}
