
/* minilogd.c
 * 
 * A pale imitation of syslogd. Most notably, doesn't write anything
 * anywhere except possibly back to syslogd.
 * 
 */

#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>

#include "fds.h"
#include "alloc.h"

static int we_own_log;
static char **buffer;
static unsigned buflines;
static int buffer_overflow;
static int debug;
static int recvsock;

void
alarm_handler(int x)
{
	alarm(0);
	if (recvsock >= 0)
	{
		close(recvsock);
		recvsock = -1;
	}
}

void
freeBuffer()
{
	struct sockaddr_un addr;
	int     sock, conn = -1;
	unsigned i;

	bzero(&addr, sizeof(addr));
	addr.sun_family = AF_LOCAL;
	strncpy(addr.sun_path, _PATH_LOG, sizeof(addr.sun_path) - 1);

	/* wait for klogd to hit syslog */
	sleep(2);

	sock = socket(AF_LOCAL, SOCK_DGRAM, 0);
	if (sock >= 0)
		conn = connect(sock, (struct sockaddr *) &addr, sizeof(addr));

	for (i = 0; i < buflines && buffer[i]; ++i)
	{
		if (conn >= 0)
			write_loop(sock, buffer[i], strlen(buffer[i]) + 1);
		free(buffer[i]);
	}
}

void
cleanup(int exitcode)
{
	/* If we own the log, unlink it before trying to free our buffer.
	 * Otherwise, sending the buffer to /dev/log doesn't make much sense.... */
	if (we_own_log)
	{
		perror("wol");
		unlink(_PATH_LOG);
	}
	/* Don't try to free buffer if we were called from a signal handler */
	if (exitcode <= 0)
	{
		if (buffer)
			freeBuffer();
		exit(exitcode);
	} else
		exit(exitcode + 128);
}

void
runDaemon(int sock)
{
	struct sockaddr_un addr;
	int     x, len, done = 0;
	socklen_t addrlen;
	struct stat s1, s2;
	struct pollfd pfds;

	if (daemon(0, -1) < 0)
	{
		perror("daemon");
		return;
	}
	/* try not to leave stale sockets lying around */
	/* Hopefully, we won't actually get any of these */
	signal(SIGHUP, cleanup);
	signal(SIGINT, cleanup);
	signal(SIGQUIT, cleanup);
	signal(SIGILL, cleanup);
	signal(SIGABRT, cleanup);
	signal(SIGFPE, cleanup);
	signal(SIGSEGV, cleanup);
	signal(SIGPIPE, cleanup);
	signal(SIGBUS, cleanup);
	signal(SIGTERM, cleanup);
	/* Get stat info on /dev/log so we can later check to make sure we
	 * still own it... */
	if (stat(_PATH_LOG, &s1) < 0)
		memset(&s1, '\0', sizeof(struct stat));
	while (!done)
	{
		pfds.fd = sock;
		pfds.events = POLLIN | POLLPRI;
		if (((x = poll(&pfds, 1, 500)) == -1) && errno != EINTR)
		{
			perror("poll");
			cleanup(-1);
		}
		if ((x > 0) && pfds.revents & (POLLIN | POLLPRI))
		{
			char    message[8192];

			memset(message, 0, sizeof(message));
			recvsock =
				accept(sock, (struct sockaddr *) &addr,
				       &addrlen);
			alarm(2);
			signal(SIGALRM, alarm_handler);
			len = read(recvsock, message, sizeof(message) - 2);
			alarm(0);
			if (recvsock >= 0)
			{
				close(recvsock);
				recvsock = -1;
			}
			if (len > 0)
				len = strlen(message);
			for (; len >= 0; --len)
			{
				if (!message[len] || isspace(message[len]))
					continue;
				message[++len] = '\n';
				message[++len] = '\0';
				break;
			}
			if (!buffer_overflow && len > 0)
			{
				char   *m = strdup(message);

				if (!m)
					buffer_overflow = 1;
				else
				{
					if (buffer)
					{
						char  **p = realloc(buffer,
								    (buflines
								     +
								     1) *
								    sizeof
								    (char *));
						if (!p)
							buffer_overflow = 1;
						else
							buffer = p;
					} else
						buffer = xmalloc(sizeof
								 (char *));

					if (!buffer_overflow)
						buffer[buflines++] = m;
				}
			}
		}
		if ((x > 0) && (pfds.revents & (POLLHUP | POLLNVAL)))
			done = 1;
		/* Check to see if syslogd's yanked our socket out from under us */
		if ((stat(_PATH_LOG, &s2) < 0) ||
		    (s1.st_ino != s2.st_ino) || (s1.st_ctime != s2.st_ctime)
		    || (s1.st_mtime != s2.st_mtime))
		{
			done = 1;
			we_own_log = 0;
		}
	}
	cleanup(0);
}

int
main(int argc, char **argv)
{
	struct sockaddr_un addr;
	int     i, fd, sock;
	int     pid;

	/* option processing made simple... */
	if (argc > 1)
		debug = 1;

	/* just in case */
	fd = open(_PATH_DEVNULL, O_RDWR);
	if (fd < 0)
		error(EXIT_FAILURE, errno, "open: %s", _PATH_DEVNULL);
	for (i = 0; i <= 2; ++i)
		if (fd != i)
			dup2(fd, i);
	if (fd > 2)
		close(fd);

	bzero(&addr, sizeof(addr));
	addr.sun_family = AF_LOCAL;
	strncpy(addr.sun_path, _PATH_LOG, sizeof(addr.sun_path) - 1);
	sock = socket(AF_LOCAL, SOCK_STREAM, 0);
	unlink(_PATH_LOG);
	/* Bind socket before forking, so we know if the server started */
	umask(0);
	if (!bind(sock, (struct sockaddr *) &addr, sizeof(addr)))
	{
		we_own_log = 1;
		listen(sock, 5);
		if ((pid = fork()) == -1)
		{
			perror("fork");
			exit(3);
		}
		if (pid)
		{
			exit(0);
		} else
		{
			runDaemon(sock);
			/* shouldn't get back here... */
			exit(4);
		}
	} else
	{
		exit(5);
	}
}
